{#-
 # Copyright (c) 2019 UAVCAN Consortium
 # This software is distributed under the terms of the MIT License.
 # Author: Pavel Kirienko <pavel@uavcan.org>
-#}
# AUTOGENERATED, DO NOT EDIT.
#
# Generated at:  {{ now_utc }} UTC
# Namespace:     {{ T.full_name }}
{% if T.get_nested_types() %}
    {%- for t, _ in T.get_nested_types() %}
from {{ t|full_reference_name }} import {{ t|short_reference_name }} as {{ t|short_reference_name }}
    {%- endfor %}

# Convenience aliases of the newest minor versions known at the time of code generation.
    {%- for alias, t in T.get_nested_types()|map("first")|newest_minor_version_aliases %}
{{ alias }} = {{ t|short_reference_name }}
    {%- endfor %}
{%- else %}
# There are no data types in this namespace. There may be data types in nested namespaces.
{%- endif %}
