#!/usr/bin/env python3
#
# Copyright (C) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import setuptools

setuptools.setup()
