{#-
 # Copyright (c) 2019 UAVCAN Consortium
 # This software is distributed under the terms of the MIT License.
 # Author: Pavel Kirienko <pavel@uavcan.org>
-#}

{% extends "base.j2" %}
{%- block contents -%}

{{ data_schema(T|short_reference_name, T) }}

{%- endblock -%}
